#ifndef BEP_HPP_
#define BEP_HPP_

#include "bhash.hpp"

namespace bep_tool {

template<class valType>
class bep {
	typedef std::map<std::string, valType> umap;
	typedef typename std::map<std::string, valType>::iterator umap_iterator;

public:
	bep()
	{
	}
	~bep()
	{
	}

	// build using string as a key and valType as a value
	bool build(const std::vector<std::string>& keys, const std::vector<valType>& vals) {
		const size_t n = vals.size();
		if (keys.size() != n) {
			fprintf(stderr, "keys.size():%u != vals.size():%u\n", (unsigned int)keys.size(), (unsigned int)n);
			return false;
		}
		if (!bh_.build(keys)) {
			return false;
		}
		vals_.resize(n);
		for (size_t i = 0; i < n; i++) {
			vals_[bh_.lookup_wocheck(keys[i])] = vals[i];
		}
		return true;
	}

	bool build(const std::vector<std::pair<std::string, valType> >& dic)
	{
		std::vector<std::string> keys;
		std::vector<valType> vals;
		size_t n = dic.size();
		keys.resize(n);
		vals.resize(n);
		for (size_t i = 0; i < n; i++) {
			keys[i] = dic[i].first;
			vals[i] = dic[i].second;
		}
		return build(keys, vals);
	}

	bool build()
	{
		std::vector<std::string> keys;
		std::vector<valType> vals;
		for (size_t i = 0; i < bh_.size(); i++) {
			std::string key;
			 bh_.lookupKey(i, key);
			 keys.push_back(key);
			 vals.push_back(vals_[i]);
		}
		for (umap_iterator it = work_.begin(), e = work_.end(); it != e; ++it) {
			keys.push_back(it->first);
			vals.push_back(it->second);
		}
		work_.clear();
		return build(keys, vals);
	}

	// save current perfect hash function to fileName
	// NOT save vals because we don't know how to store vals
	bool save(const char *fileName, std::vector<valType>& vals)
	{
		if (!build()) return false;
		vals = vals_;
		return bh_.save(fileName);
	}
	// load fileName to current perfect hash function
	bool load(const char *fileName, const std::vector<valType>& vals)
	{
		if (!bh_.load(fileName)) return false;
		const size_t n = bh_.size();
		vals_.resize(n);
		for (size_t i = 0; i < n; i++) {
			vals_[i] = vals[i];
		}
		return true;
	}

	// return registered val
	valType& operator[](const std::string& key)
	{
		uint32 id = 0;
		if (bh_.size() > 0 && (id = bh_.lookup(key)) != NOTFOUND) {
			return vals_[id];
		}
		umap_iterator it = work_.find(key);
		if (it != work_.end()) {
			return it->second;
		} else {
			const size_t MAXWORKSIZE = 1048576;	// 1M
			if (work_.size() == MAXWORKSIZE) {
				build();	// merge work_ and bh_
			}
			return work_[key];
		}
	}

	bool exist(const std::string& key) const
	{
		if (bh_.lookup(key) == NOTFOUND) {
			if (work_.count(key) == 0) return false;
		}
		return true;
	}

	size_t size() const { return bh_.size() + work_.size(); }
private:
	bool lookup(const size_t i, std::string& key, valType& val) const
	{
		if (i >= bh_.size()) return false;
		bh_.lookupKey(i, key);
		val = vals_[i];
		return true;
	}

	bhash bh_; // perfect hash. key to ID
	std::vector<valType> vals_;
	umap work_;
};

} // end of namespace

#endif

