#include "bhash.hpp"
#include <fstream>
#include <queue>
#include <map>

using namespace bep_tool;

static const char *KEYEXT = ".key";
static const double C_R = 1.3;

HashFunctionType *bep_tool::s_hashFunction = 0;

// Hash function: http://www.burtleburtle.net/bob/c/lookup8.c
/* use macro for VC(faster) */
#ifdef BEP_USE_LIGHTWEIGHT_HASH
#define bob_mix(a, b, c) \
	a -= b; a -= c; a ^= (c >> 13); \
	b -= c; b -= a; b ^= (a << 8); \
	c -= a; c -= b; c ^= (b >> 13); \
	a -= b; a -= c; a ^= (c >> 12);
#else
#define bob_mix(a, b, c) \
	a -= b; a -= c; a ^= (c >> 13); \
	b -= c; b -= a; b ^= (a << 8); \
	c -= a; c -= b; c ^= (b >> 13); \
	a -= b; a -= c; a ^= (c >> 12); \
	b -= c; b -= a; b ^= (a << 16); \
	c -= a; c -= b; c ^= (b >> 5); \
	a -= b; a -= c; a ^= (c >> 3); \
	b -= c; b -= a; b ^= (a << 10); \
	c -= a; c -= b; c ^= (b >> 15);
#endif

static inline uint32 get32bitAsLE(const uint8* p)
{
	return p[0] | (p[1] << 8) | (p[2] << 16) | (p[3] << 24);
}

static inline void defaultHash(const std::string& str, const uint32 init, uint32* pa, uint32* pb, uint32* pc)
{
	const uint8* p = reinterpret_cast<const uint8*>(str.c_str());
	size_t len = str.size();
	uint32 a, b, c;
	a = b = 0x9e3779b9; c = init;

	while (len >= 12) {
		a += get32bitAsLE(p + 0);
		b += get32bitAsLE(p + 4);
		c += get32bitAsLE(p + 8);
		bob_mix(a, b, c);
		p += 12;
		len -= 12;
	}
	c += static_cast<uint32>(str.size());
	switch (len) {
	case 11: c += p[10] << 24;
	case 10: c += p[9] << 16;
	case 9: c += p[8] << 8;
		// the first byte of c is reserved for the length
	case 8: b += p[7] << 24;
	case 7: b += p[6] << 16;
	case 6: b += p[5] << 8;
	case 5: b += p[4];
	case 4: a += p[3] << 24;
	case 3: a += p[2] << 16;
	case 2: a += p[1] << 8;
	case 1: a += p[0];
	}
	bob_mix(a, b, c);
	*pa = a; *pb = b; *pc = c;
}

bool bhash::build(const std::vector<std::string>& keys, HashFunctionType *func)
{
	s_hashFunction = func ? func : defaultHash;
	if (keys.size() >= 0x80000000L / C_R) return false;
	keyNum_ = static_cast<uint32>(keys.size());
	bucketNum_ = (uint32)(keyNum_ * C_R / 3.0);
	totalBucketNum_ = bucketNum_ * 3;

	std::vector<uint32> deg(totalBucketNum_);
	std::vector<uint32> edges(keyNum_ * 3);
	std::vector<uint32> offset(totalBucketNum_ + 1);
	std::vector<bool> visitedEdges(keyNum_);
	std::vector<bool> visitedVerticies(totalBucketNum_);
	std::vector<uint8> g(totalBucketNum_);		// vertex marker

	std::vector<hg_edge> vs;
	seed_ = 0x12345678;
	bool edgeSuccess = true;
RETRY:
	for (int iter = 0; iter < 20; iter++) {
		seed_++;
		std::fill(deg.begin(), deg.end(), 0);

		// generate candidate edges
		vs.resize(keyNum_);
		for (size_t i = 0; i < keyNum_; i++) {
			hg_edge v(keys[i], seed_, bucketNum_);
			vs[i] = v;
			deg[v.v[0]]++;
			deg[v.v[1]]++;
			deg[v.v[2]]++;
		}

		// check whether identical edges exist or not
		sort(vs.begin(), vs.end());

		for (size_t i = 1, n = vs.size(); i < n; i++) {
			if (vs[i - 1] == vs[i]) {
				fprintf(stderr, "find identical edge[%d]\n", iter);
				edgeSuccess = false;
				break;
			}
		}
		if (edgeSuccess) break;
	}
	if (!edgeSuccess) {
		fprintf(stderr, "retired\n");
		return false;
	}

	// build offset and edges
	uint32 sum = 0;
	for (uint32 i = 0; i < totalBucketNum_; i++) {
		offset[i] = sum;
		sum += deg[i];
	}
	offset[totalBucketNum_] = sum;
	std::fill(deg.begin(), deg.end(), 0);

	for (uint32 i = 0; i < keyNum_; i++) {
		for (uint32 j = 0; j < 3; j++) {
			uint32 t = vs[i].v[j];
			edges[offset[t] + deg[t]] = i;
			deg[t]++;
		}
	}

	// check validity

	// initialize
	const uint8 unmarked = 3; /* don't change this value */
	std::fill(visitedEdges.begin(), visitedEdges.end(), false);
	std::fill(visitedVerticies.begin(), visitedVerticies.end(), false);
	std::fill(g.begin(), g.end(), unmarked);

	std::queue<uint32> q;
	for (size_t i = 0; i < totalBucketNum_; i++) {
		if (deg[i] == 1) q.push(edges[offset[i]]);
	}

	uint32 deleteNum = 0;
	printf("n:%u bn:%u\n", keyNum_, totalBucketNum_);

	std::vector<std::pair<uint32, uint32> > extractedEdges;

	while (!q.empty()) {
		uint32 eID = q.front();
		q.pop(); // edgeID

		if (visitedEdges[eID]) continue;
		deleteNum++;
		visitedEdges[eID] = true;

		const hg_edge& e = vs[eID];
		int choosed = -1;
		for (int j = 0; j < 3; j++) {
			const uint32 t = e.v[j];
			deg[t]--;
			if (deg[t] == 1) {
				for (uint32 i = offset[t]; i < offset[t + 1]; i++) {
					if (!visitedEdges[edges[i]]) {
						q.push(edges[i]);
						break;
					}
				}
			} else if (deg[t] == 0) {
				choosed = j;
			}
		}

		if (choosed == -1) {
			fprintf(stderr, "unexpected error: we cannot find free vertex\n");
			break;
		}
		extractedEdges.push_back(std::make_pair(eID, choosed));
		//printf("choose %u set %u\n", e.v[choosed], val%3);
	}

	if (deleteNum != keyNum_) goto RETRY;
	// Current candidate can generate perfect functions
	reverse(extractedEdges.begin(), extractedEdges.end());
	for (size_t i = 0, n = extractedEdges.size(); i < n; i++) {
		const hg_edge& e = vs[extractedEdges[i].first];
		const uint32 choosed = extractedEdges[i].second;
		uint32 val = choosed + 30;		// +30: large offsets for limiting positive value

		for (int j = 0; j < 3; j++) {
			if (!visitedVerticies[e.v[j]]) {
				continue;
			}
			val -= g[e.v[j]];
		}
		g[e.v[choosed]] = static_cast<uint8>(val % 3);
		visitedVerticies[e.v[choosed]] = true;
	}

	// set rank table and gTbl_
	levelA_.resize(GetNum(totalBucketNum_, 256));
	levelB_.resize(GetNum(totalBucketNum_, INTBLOCK));
	gTbl_.init(totalBucketNum_);

	uint32 r = 0;
	for (uint32 i = 0; i < totalBucketNum_; i++) {
		if (i % 256 == 0) levelA_[i / 256] = r;
		if (i % INTBLOCK == 0) levelB_[i / INTBLOCK] = static_cast<uint8>(r - levelA_[i / 256]);
		if (g[i] != unmarked) {
			r++;
		}
		gTbl_.set(i, g[i]); /* share B in original source */
	}

	// set original keys 
	totalKeySize_ = 0;
	std::vector<uint32> hashedTbl(keyNum_);
	for (uint32 i = 0; i < keyNum_; i++) {
		totalKeySize_ += keys[i].size();
		hashedTbl[lookup_wocheck(keys[i])] = i;
	}
	strTbl_.resize(totalKeySize_);
	strOffset_.resize(keyNum_ + 1);

	DataSize lenSum = 0;
	for (uint32 i = 0; i < keyNum_; i++) {
		strOffset_[i] = lenSum;
		const std::string& str = keys[hashedTbl[i]];
		memcpy(&strTbl_[lenSum], &str[0], str.size());
		lenSum += str.size();
	}
	strOffset_[keyNum_] = lenSum;
	return true;
}

template<typename T>
void writeVector(std::ofstream& ofs, const T& in)
{
	ofs.write((const char*)&in[0], in.size() * sizeof(in[0]));
}

bool bhash::save(const char *fileName) const {
	std::ofstream file(fileName, std::ios::binary);
	if (!file) return false;
	std::string fileNameKey(fileName);
	fileNameKey += KEYEXT;
	std::ofstream key(fileNameKey.c_str(), std::ios::binary);
	if (!key) return false;

//printf("save n=%d, seed=%d, totalKeySize=%d\n", keyNum_, seed_, totalKeySize_);
	file << keyNum_ << std::endl << seed_ << std::endl;
	key << totalKeySize_ << std::endl;
	writeVector(file, levelA_);
	writeVector(file, levelB_);
	file.write((const char*)gTbl_.getRawPointer(), gTbl_.getSize());
	writeVector(key, strTbl_);
	writeVector(key, strOffset_);
	return true;
}

template<typename T>
void resizeAndRead(std::ifstream& ifs, T& out, size_t n)
{
	out.resize(n);
	ifs.read((char*)&out[0], sizeof(out[0]) *n);
}
bool bhash::load(const char *fileName)
{
	std::ifstream file(fileName, std::ios::binary);
	if (!file) return false;
	std::string fileNameKey(fileName);
	fileNameKey += KEYEXT;
	std::ifstream key(fileNameKey.c_str(), std::ios::binary);
	if (!key) return false;
	file >> keyNum_ >> seed_;
	file.get(); // skip std::endl
	key >> totalKeySize_;
	key.get(); // skip std::endl
	bucketNum_ = static_cast<uint32>(keyNum_ * C_R / 3.0);
	totalBucketNum_ = bucketNum_ * 3;

//printf("load n=%d, seed=%d, totalKeySize=%d\n", keyNum_, seed_, totalKeySize_);

	resizeAndRead(file, levelA_, GetNum(totalBucketNum_, 256));
	resizeAndRead(file, levelB_, GetNum(totalBucketNum_, INTBLOCK));
	gTbl_.init(totalBucketNum_);
	file.read((char*)gTbl_.getRawPointer(), gTbl_.getSize());
	resizeAndRead(key, strTbl_, totalKeySize_);
	resizeAndRead(key, strOffset_, keyNum_ + 1);
	return true;
}

