#ifndef BHASH_HPP_
#define BHASH_HPP_

#include <string>
#include <vector>
#include <memory.h>
#include <cassert>

#if defined(_MSC_VER) && (_MSC_VER >= 1400) && (_SECURE_SCL == 1)
	#pragma message("define _SECURE_SCL = 0 for speed up")
#endif

//#define BEP_OVER_4GB_DATA
//#define BEP_USE_LIGHTWEIGHT_HASH

namespace bep_tool {

#ifdef __linux__
typedef unsigned long long uint64;
#else
typedef unsigned __int64 uint64;
#endif
typedef unsigned int uint32;
typedef unsigned char uint8;

#ifdef BEP_OVER_4GB_DATA
	typedef uint64 DataSize; /* 8byte if 64bit OS */
#else
	typedef uint32 DataSize; /* 32bit */
#endif

typedef void (HashFunctionType)(const std::string&, const uint32, uint32*, uint32*, uint32*);
extern HashFunctionType *s_hashFunction;

static inline size_t GetNum(size_t size, size_t block)
{
	return (size + block - 1) / block;
}

/* 2bit vector */
class Vect2bit {
	std::vector<uint32> p_;
	enum { n = 16 };
public:
	void init(uint32 bitLen)
	{
		p_.resize(GetNum(bitLen, n));
		std::fill(p_.begin(), p_.end(), 0);
	}
	inline void set(uint32 pos, uint32 val)
	{
		uint32 q = pos / n;
		uint32 r = (pos % n) * 2;
		p_[q] |= val << r;
	}
	inline uint32 get(uint32 pos) const
	{
		uint32 q = pos / n;
		uint32 r = (pos % n) * 2;
		return (p_[q] >> r) & 3;
	}
	uint32 *getRawPointer() { return &p_[0]; }
	const uint32 *getRawPointer() const { return &p_[0]; }
	size_t getSize() const { return p_.size() * sizeof(p_[0]); }
};

struct hg_edge {
	uint32 v[3];
public:
	hg_edge()
	{
		v[0] = v[1] = v[2] = 0;
	}
	bool operator<(const hg_edge& rhs) const
	{
		return v[0] != rhs.v[0] ? v[0] < rhs.v[0] : v[1] != rhs.v[1] ? v[1] < rhs.v[1] : v[2] < rhs.v[2];
	}
	bool operator==(const hg_edge& rhs) const
	{
		return ((v[0] ^ rhs.v[0]) | (v[1] ^ rhs.v[1]) | (v[2] ^ rhs.v[2])) == 0;
	}
	hg_edge(const std::string& str, uint32 seed, uint32 n)
	{
		s_hashFunction(str, seed, &v[0], &v[1], &v[2]);
		v[0] = (v[0] % n);
		v[1] = (v[1] % n) + n;
		v[2] = (v[2] % n) + n * 2;
	}
};

static const uint32 NOTFOUND = 0xffffffff;

class bhash {
	enum {
		INTBLOCK = 32
	};
public:
	// rank tables
	inline uint32 rank(const uint32 i) const
	{
		return levelA_[i / 256] + levelB_[i / INTBLOCK] + popCount2(gTbl_.getRawPointer() + (i / 16),  i % INTBLOCK);
	}
	// count gTbl_ directory(unmarked is unused)
	inline uint32 popCount2(const uint32 *p, uint32 len) const
	{
		static const uint32 mask[16] = {
			0x00000000, 0x00000003, 0x0000000F, 0x0000003F,
			0x000000FF,	0x000003FF, 0x00000FFF, 0x00003FFF,
			0x0000FFFF, 0x0003FFFF, 0x000FFFFF, 0x003FFFFF,
			0x00FFFFFF, 0x03FFFFFF, 0x0FFFFFFF, 0x3FFFFFFF,
		};
		if (len < 16) {
			uint32 x = (~p[0]) & mask[len];
			x = ((x >> 1) | x) & 0x55555555;
			return popCount(x);
		} else {
			uint64 x = ((uint64)(~p[-1]) << 32) | ((~p[0]) & mask[len - 16]);
			x = ((x >> 1) | x) & 0x5555555555555555LL;
			return popCount(uint32(x >> 32) | ((uint32(x) << 1)));
		}
	}
	inline uint32 popCount(uint32 x) const
	{
#if 1 // a little faster
		static const uint8 tbl[] = {
			0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
			1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
			1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
			2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
			1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
			2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
			2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
			3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
			1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
			2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
			2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
			3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
			2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
			3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
			3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
			4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8,
		};
		return tbl[x & 0xff] + tbl[(x >> 8) & 0xff] + tbl[(x >> 16) & 0xff] + tbl[x >> 24];
#else
		x = ((x >> 1) & 0x55555555) + (x & 0x55555555);
		x = ((x >> 2) & 0x33333333) + (x & 0x33333333);
		x = ((x >> 4) + x) & 0x0F0F0F0F;
		x = (x >> 8) + x;
		x = ((x >> 16) + x) & 0x3F;
		return x;
#endif
	}
	DataSize totalKeySize_; // total length of keys
	std::vector<DataSize> strOffset_; // store key's offsets
	std::vector<uint32> levelA_; // store every 256-th rank result
	std::vector<uint8> levelB_; // store every  32-th rank result
	Vect2bit gTbl_; // store g values in compressed representations
	std::vector<char> strTbl_; // store original keys in raw format
	uint32 seed_; // seed_ for hash functions
	uint32 keyNum_; // number of keys
	uint32 bucketNum_; // number of buckets for each hash function
	uint32 totalBucketNum_; // number of total buckets (output of NON-minimal perfect hashing)
public:
	bhash()
		: totalKeySize_(0)
		, seed_(0)
		, keyNum_(0)
		, bucketNum_(0)
		, totalBucketNum_(0)
	{
	}
	~bhash()
	{
	}
	bool build(const std::vector<std::string>& keys, HashFunctionType *func = 0);	// build perfect hash functions for keys
	bool save(const char *fileName) const;	// save current perfect hash function to fileName
	bool load(const char *fileName);		// load fileName to current perfect hash function
	// lookup key and return ID or NOTFOUND
	uint32 lookup(const std::string& key) const
	{
		const uint32 id = lookup_wocheck(key);
		const DataSize begin = strOffset_[id];
		const DataSize len = strOffset_[id + 1] - begin;
		if (key.size() != len) return NOTFOUND;
		if (memcmp(key.c_str(), &strTbl_[begin], key.size()) != 0) return NOTFOUND;
		return id;
	}
	// lookup key and may return NOTFOUND
	// lookup key and return ID without checking unknown keys
	uint32 lookup_wocheck(const std::string& key) const
	{
		hg_edge v(key, seed_, bucketNum_);
		uint32 val = gTbl_.get(v.v[0]);
		val += gTbl_.get(v.v[1]);
		val += gTbl_.get(v.v[2]);
		return rank(v.v[val % 3]);
	}
	// return key which retruns i as ID
	bool lookupKey(const uint32 i, std::string& key) const
	{
		if (i >= keyNum_) return false;
		key.assign(&strTbl_[strOffset_[i]], strOffset_[i + 1] - strOffset_[i]);
		return true;
	}
	uint32 size() const { return keyNum_; }
};

} // bep_tool

#endif
